/*
 *                           TERMS AND CONDITIONS
 *                                   FOR
 *                         OPEN SOURCE CODE LICENSE
 *                               Version 1.1
 * 
 * Japan Registry Services Co., Ltd. ("JPRS"), a Japanese corporation
 * having its head office at Chiyoda First Bldg. East 13F 3-8-1 Nishi-Kanda,
 * Chiyoda-ku, Tokyo 101-0065, Japan, grants you the license for open source
 * code specified in EXHIBIT A the "Code" subject to the following Terms and
 * Conditions ("OSCL").
 * 
 * 1. License Grant.
 *   JPRS hereby grants you a worldwide, royalty-free, non-exclusive
 *   license, subject to third party intellectual property claims:
 *   (a) under intellectual property rights (other than patent or
 *       trademark) licensable by JPRS to use, reproduce, modify, display,
 *       perform, sublicense and distribute the Code (or portions thereof)
 *       with or without modifications, and/or as part of a derivative work;
 *       or
 *   (b) under claims of the infringement through the making, using,
 *       offering to sell and/or otherwise disposing the JPRS Revised Code
 *       (or portions thereof);
 *   (c) the licenses granted in this Section 1(a) and (b) are effective on
 *       the date JPRS first distributes the Code to you under the terms of
 *       this OSCL;
 *   (d) Notwithstanding the above stated terms, no patent license is
 *       granted:
 *       1)  for a code that you delete from the Code;
 *       2)  separate from the Code; or
 *       3)  for infringements caused by:
 *            i) modification of the Code; or
 *           ii) combination of the Code with other software or devices.
 * 
 * 2. Consents.
 *   You agree that:
 *   (a) you must include a copy of this OSCL and the notice set forth in
 *       EXHIBIT A with every copy of the Code you distribute;
 *   (b) you must include a copy of this OSCL and the notice set forth in
 *       EXHIBIT A with every copy of binary form of the Code in the
 *       documentation and/or other materials provided with the distribution;
 *   (c) you may not offer or impose any terms on any source code version
 *       that alters or restricts the applicable version of this OSCL or
 *       the recipients' rights hereunder.
 *   (d) If the terms and conditions are set forth in EXHIBIT A, you must
 *       comply with those terms and conditions.
 * 
 * 3. Proprietary Information.
 *   All trademarks, service marks, patents, copyrights, trade secrets, and
 *   other proprietary rights in or related to the Code are and will remain
 *   the exclusive property of JPRS or its licensors, whether or not
 *   specifically recognized or perfected under local law except specified
 *   in this OSCL; provided however you agree and understand that the JPRS
 *   name may not be used to endorse or promote this Code without prior
 *   written approval of JPRS.
 * 
 * 4. WARRANTY DISCLAIMER.
 *   JPRS MAKES NO REPRESENTATIONS AND WARRANTIES REGARDING THE USE OF THE
 *   CODE, NOR DOES JPRS MAKE ANY REPRESENTATIONS THAT THE CODE WILL BECOME
 *   COMMERCIALLY AVAILABLE. JPRS, ITS AFFILIATES, AND ITS SUPPLIERS DO NOT
 *   WARRANT OR REPRESENT THAT THE CODE IS FREE OF ERRORS OR THAT THE CODE
 *   IS SUITABLE FOR TRANSLATION AND/OR LOCALIZATION. THE CODE IS PROVIDED
 *   ON AN "AS IS" BASIS AND JPRS AND ITS SUPPLIERS HAVE NO OBLIGATION TO
 *   CORRECT ERRORS OR TO SUPPORT THE CODE UNDER THIS OSCL FOR ANY REASON.
 *   TO THE FULL EXTENT PERMITTED BY LAW, ALL OBLIGATIONS ARE HEREBY
 *   EXCLUDED WHETHER EXPRESS, STATUTORY OR IMPLIED UNDER LAW, COURSE OF
 *   DEALING, CUSTOM, TRADE USAGE, ORAL OR WRITTEN STATEMENT OR OTHERWISE,
 *   INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY
 *   OR FITNESS FOR A PARTICULAR PURPOSE CONCERNING THE CODE.
 * 
 * 5. NO LIABILITY.
 *   UNDER NO CIRCUMSTANCES SHALL JPRS AND/OR ITS AFFILIATES, LICENSORS, OR
 *   REPRESENTATIVES BE LIABLE FOR ANY DAMAGES INCLUDING BUT NOT LIMITED TO
 *   CONSEQUENTIAL, INDIRECT, SPECIAL, PUNITIVE OR INCIDENTAL DAMAGES,
 *   WHETHER FORESEEABLE OR UNFORESEEABLE, BASED ON YOUR CLAIMS, INCLUDING,
 *   BUT NOT LIMITED TO, CLAIMS FOR LOSS OF DATA, GOODWILL, PROFITS, USE OF
 *   MONEY, INTERRUPTION IN USE OR AVAILABILITY OF DATA, STOPPAGE, IMPLIED
 *   WARRANTY, BREACH OF CONTRACT, MISREPRESENTATION, NEGLIGENCE, STRICT
 *   LIABILITY IN TORT, OR OTHERWISE.
 * 
 * 6. Indemnification.
 *   You hereby agree to indemnify, defend, and hold harmless JPRS for any
 *   liability incurred by JRPS due to your terms of warranty, support,
 *   indemnity, or liability offered by you to any third party.
 * 
 * 7. Termination.
 * 7.1 This OSCL shall be automatically terminated in the events that:
 *   (a) You fail to comply with the terms herein and fail to cure such
 *       breach within 30 days of becoming aware of the breach;
 *   (b) You initiate patent or copyright infringement litigation against
 *       any party (including a cross-claim or counterclaim in a lawsuit)
 *       alleging that the Code constitutes a direct or indirect patent or
 *       copyright infringement, in such case, this OSCL to you shall
 *       terminate as of the date such litigation is filed;
 * 7.2 In the event of termination under Sections 7.1(a) or 7.1(b) above,
 *     all end user license agreements (excluding distributors and
 *     resellers) which have been validly granted by You or any distributor
 *     hereunder prior to termination shall survive termination.
 *
 * 
 * 8. General.
 *   This OSCL shall be governed by, and construed and enforced in
 *   accordance with, the laws of Japan. Any litigation or arbitration
 *   between the parties shall be conducted exclusively in Tokyo, Japan
 *   except written consent of JPRS provides other venue.
 * 
 * 
 *                                EXHIBIT A
 * 
 * The original open source code of idnkit-2 is idnkit-1.0 developed and
 * conceived by Japan Network Information Center ("JPNIC"), a Japanese
 * association, Kokusai-Kougyou-Kanda Bldg 6F, 2-3-4 Uchi-Kanda,
 * Chiyoda-ku, Tokyo 101-0047, Japan, and JPRS modifies above original code
 * under following Terms and Conditions set forth by JPNIC.
 * 
 *                                  JPNIC
 * 
 * Copyright (c) 2000-2002 Japan Network Information Center.  All rights reserved.
 * 
 * By using this file, you agree to the terms and conditions set forth bellow.
 * 
 *                       LICENSE TERMS AND CONDITIONS
 * 
 * The following License Terms and Conditions apply, unless a different
 * license is obtained from Japan Network Information Center ("JPNIC"),
 * a Japanese association, Kokusai-Kougyou-Kanda Bldg 6F, 2-3-4 Uchi-Kanda,
 * Chiyoda-ku, Tokyo 101-0047, Japan.
 * 
 * 1. Use, Modification and Redistribution (including distribution of any
 *    modified or derived work) in source and/or binary forms is permitted
 *    under this License Terms and Conditions.
 * 
 * 2. Redistribution of source code must retain the copyright notices as they
 *    appear in each source code file, this License Terms and Conditions.
 * 
 * 3. Redistribution in binary form must reproduce the Copyright Notice,
 *    this License Terms and Conditions, in the documentation and/or other
 *    materials provided with the distribution. For the purposes of binary
 *    distribution the "Copyright Notice" refers to the following language:
 *    "Copyright (c) 2000-2002 Japan Network Information Center.  All rights reserved."
 * 
 * 4. The name of JPNIC may not be used to endorse or promote products
 *    derived from this Software without specific prior written approval of
 *    JPNIC.
 * 
 * 5. Disclaimer/Limitation of Liability: THIS SOFTWARE IS PROVIDED BY JPNIC
 *    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL JPNIC BE LIABLE
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *    BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *    WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *    OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *    ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * 
 *                        JPRS Public License Notice
 *                                   For
 *                                idnkit-2.
 * 
 * The contents of this file are subject to the Terms and Conditions for
 * the Open Source Code License (the "OSCL"). You may not use this file
 * except in compliance with above terms and conditions. A copy of the OSCL
 * is available at <http://jprs.co.jp/idn/>.
 * The JPRS Revised Code is idnkit-2.
 * The Initial Developer of the JPRS Revised Code is Japan Network
 * Information Center ("JPNIC"), a Japanese association,
 * Kokusai-Kougyou-Kanda Bldg 6F, 2-3-4 Uchi-Kanda, Chiyoda-ku, Tokyo
 * 101-0047, Japan.
 * "Copyright (c) 2000-2002 Japan Network Information Center.  All rights reserved."
 * "Copyright (c) 2010-2012 Japan Registry Services Co., Ltd.  All rights reserved."
 * Contributor(s): ______________________________________.
 * 
 * If you wish to allow use of your version of this file only under the
 * above License(s) and not to allow others to use your version of this
 * file, please indicate your decision by deleting the relevant provisions
 * above and replacing them with the notice and other provisions required
 * by the above License(s). If you do not delete the relevant provisions,
 * a recipient may use your version of this file under either the above
 * License(s).
 */

#include <config.h>

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <idn/assert.h>
#include <idn/debug.h>
#include <idn/logmacro.h>
#include <idn/result.h>
#include <idn/util.h>

/*
 * ASCII ctype macros.
 * Note that these macros evaluate the argument multiple times.  Be careful.
 */
#define ASCII_TOUPPER(c) \
	(('a' <= (c) && (c) <= 'z') ? ((c) - 'a' + 'A') : (c))
#define ASCII_TOLOWER(c) \
	(('A' <= (c) && (c) <= 'Z') ? ((c) - 'A' + 'a') : (c))

/*
 * Copy a string 'from' to 'to'.
 */
idn_result_t
idn__util_strcpy(char *to, size_t tolen, const char *from) {
	while (*from != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		*to++ = *from++;
		tolen--;
	}

	if (tolen <= 0)
		return (idn_buffer_overflow);
	*to = '\0';

	return (idn_success);
}

/*
 * Copy at most first 'n' bytes of a string 'from' to 'to'.
 */
idn_result_t
idn__util_strncpy(char *to, size_t tolen, const char *from, size_t n) {
	while (n > 0 && *from != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		*to++ = *from++;
		tolen--;
		n--;
	}

	if (tolen <= 0)
		return (idn_buffer_overflow);
	*to = '\0';

	return (idn_success);
}

/*
 * Append a string 'from' to 'to'.
 */
idn_result_t
idn__util_strcat(char *to, size_t tolen, const char *from) {
	while (*to != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		to++;
		tolen--;
	}

	while (*from != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		*to++ = *from++;
		tolen--;
	}

	if (tolen <= 0)
		return (idn_buffer_overflow);
	*to = '\0';

	return (idn_success);
}

/*
 * Append at most first 'n' bytes of a string 'from' to 'to'.
 */
idn_result_t
idn__util_strncat(char *to, size_t tolen, const char *from, size_t n) {
	while (*to != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		to++;
		tolen--;
	}

	while (n > 0 && *from != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		*to++ = *from++;
		tolen--;
		n--;
	}

	if (tolen <= 0)
		return (idn_buffer_overflow);
	*to = '\0';

	return (idn_success);
}

/*
 * This function is equivalent to strdup().
 */
char *
idn__util_strdup(const char *s) {
	size_t len;
	char *result;

	assert(s != NULL);

	len = strlen(s);
	result = (char *)malloc(len + 1);
	if (result == NULL)
		return (NULL);

	memcpy(result, s, len + 1);
	return (result);
}

/*
 * Convert ASCII uppercase letters ('A'..'Z') in 's' to lowercase.
 */
void
idn__util_asclower(char *s) {
	char *p;

	assert(s != NULL);

	for (p = s; *p != '\0'; p++) {
		if ('A' <= *p && *p <= 'Z')
			*p += 'a' - 'A';
	}
}

/*
 * strcasecmp() compatible function.
 */
int
idn__util_strcasecmp(const char *str1, const char *str2) {
	unsigned long c1, c2;

	while (*str1 != '\0') {
		c1 = ASCII_TOLOWER(*str1);
		c2 = ASCII_TOLOWER(*str2);
		if (c1 > c2)
			return (1);
		else if (c1 < c2)
			return (-1);
		str1++;
		str2++;
	}

	c1 = ASCII_TOLOWER(*str1);
	c2 = ASCII_TOLOWER(*str2);
	if (c1 > c2)
		return (1);
	else if (c1 < c2)
		return (-1);

	return (0);
}

/*
 * strncasecmp() compatible function.
 */
int
idn__util_strncasecmp(const char *str1, const char *str2, size_t n) {
	unsigned long c1, c2;

	while (n > 0 && *str1 != '\0') {
		c1 = ASCII_TOLOWER(*str1);
		c2 = ASCII_TOLOWER(*str2);
		if (c1 > c2)
			return (1);
		else if (c1 < c2)
			return (-1);
		str1++;
		str2++;
		n--;
	}

	if (n > 0) {
		c1 = ASCII_TOLOWER(*str1);
		c2 = ASCII_TOLOWER(*str2);
		if (c1 > c2)
			return (1);
		else if (c1 < c2)
			return (-1);
	}

	return (0);
}

/*
 * WIN32 specific utilities.
 */
#ifdef WIN32
#include <windows.h>
#include <userenv.h>
#undef ERROR

#ifdef __WIN64__
#define REGISTRY_SUBKEY		"Software\\idnkit2 x64"
#else
#define REGISTRY_SUBKEY		"Software\\idnkit2"
#endif
#define MAX_KEY_NAME_LENGTH	255

static int	idn__util_win32getregistrystring(const char *name,
						 char *value, size_t len);

/*
 * Get user directory.
 */
int
idn__util_win32getuserdirectory(char *value, size_t len) {
	DWORD l = len;
	HANDLE hToken;

	if (!OpenProcessToken(GetCurrentProcess(), TOKEN_READ, &hToken))
		return (0);
	if (!GetUserProfileDirectory(hToken, value, &l)) {
		CloseHandle(hToken);
		return (0);
	}
	CloseHandle(hToken);
	return (1);
}

/*
 * Get SYSCONFDIR from the registry.
 */
int
idn__util_win32getsysconfdir(char *value, size_t len) {
	return idn__util_win32getregistrystring("sysconfdir", value, len);
}

/*
 * Get a value from the registry.
 */
static int
idn__util_win32getregistrystring(const char *name, char *value, size_t len) {
	static HKEY keys[] = {HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE};
	LONG stat;
	HKEY hk;
	DWORD l, type;
	int i;

	assert(name != NULL && value != NULL);

	for (i = 0; i < sizeof(keys) / sizeof(HKEY); i++) {
		stat = RegOpenKeyEx(keys[i], REGISTRY_SUBKEY, 0, KEY_READ,
				    &hk);
		if (stat != ERROR_SUCCESS)
			continue;
		l = (DWORD)len;
		stat = RegQueryValueEx(hk, (LPCTSTR)name, NULL, &type,
				       (LPBYTE)value, &l);
		RegCloseKey(hk);
		if (stat == ERROR_SUCCESS && type == REG_SZ)
			return (1);
	}

	return (0);
}

#endif /* WIN32 */
